﻿using ERNW.PowerShell.AdministrativeTemplate.AdmxSqliteProcessor;
using ERNW.PowerShell.AdministrativeTemplate.AdmxTemplate;
using System;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Management.Automation;

namespace ERNW.PowerShell.AdministrativeTemplate.PSAdmxCmdlet
{
    [Cmdlet(VerbsCommon.New, "AdmxDatabase")]
    [OutputType(typeof(CreateDBInfo))]
    public class NewAdmxDatabase : PSCmdlet
    {

        [Parameter(
            Mandatory = false,
            ValueFromPipelineByPropertyName = true,
            ValueFromPipeline = true,
            Position = 0,
            HelpMessage = "The destination path of the sqlite database"
        )]
        public string Destination { get; set; } = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments) + @"\AdmxDatabase\";

        [Parameter(
            Mandatory = false,
            ValueFromPipelineByPropertyName = true,
            ValueFromPipeline = true,
            Position = 1,
            HelpMessage = "The administrative template directory source path"
        )]
        public string Source { get; set; } = @"C:\Windows\PolicyDefinitions";


        protected override void ProcessRecord()
        {
            string baseConnectionString;

            PluginResolver.InitPluginResolver(System.IO.Path.GetDirectoryName(System.Reflection.Assembly.GetExecutingAssembly().Location));
            DirectoryInfo destinationDirInfo = Directory.CreateDirectory(Destination);

            if (destinationDirInfo.Exists)
            {
                if (!destinationDirInfo.FullName.Last().Equals('\\'))
                {
                    baseConnectionString = Path.Combine(destinationDirInfo.FullName, @"\") + DateTime.Now.ToString("s").Replace(":", "") + Environment.MachineName + ".sqlite3";
                }
                else
                    baseConnectionString = Destination + DateTime.Now.ToString("s").Replace(":", "") + Environment.MachineName + ".sqlite3";

                using (SqliteProcessor sqliteProcessor = new SqliteProcessor())
                {
                    Stopwatch sw = Stopwatch.StartNew();

                    DirectoryInfo sourceDirInfo = new DirectoryInfo(Source);
                    ProgressRecord progressBar = new ProgressRecord(1, "NewAdmxDatabase", "Progress:");


                    sqliteProcessor.CreateSqliteAdmxTable(baseConnectionString);

                    sqliteProcessor.ExecuteAdmxNonQuery($@"INSERT INTO BaseTable (base_id, base_name, base_time, base_osversion, base_count) 
                                            VALUES
                                                ('{sqliteProcessor.BaseID}', 'default', '{sqliteProcessor.BaseTime}','{Environment.OSVersion}', 0)
                                            ");

                    if (sourceDirInfo.Exists)
                    {
                        double inc = 0;
                        int basecount = 0;
                        double countPercent = (double)100 / (double)sourceDirInfo.EnumerateFiles().Count();
                        progressBar.PercentComplete = (int)inc;

                        foreach (var admxTemplateFilePath in Directory.GetFiles(Source, "*.admx", SearchOption.TopDirectoryOnly))
                        {
                            progressBar.StatusDescription = "Progress: " + admxTemplateFilePath;
                            WriteProgress(progressBar);
                            inc++;
                            var transaction = sqliteProcessor.Connection.BeginTransaction();

                            try
                            {
                                TemplateParser templateParserObject = null;
                                templateParserObject = new TemplateParser(admxTemplateFilePath, "en-US");


                                basecount += templateParserObject.TemplateObject.Count;
                                foreach (TemplateObject templateObjectPolicy in templateParserObject.TemplateObject)
                                {

                                    string jsonTemplateObjectPolicy = templateParserObject.TemplateObjectJsonSerialize<object>(templateObjectPolicy);
                                    sqliteProcessor.ExecuteAdmxNonQueryAsync($@"INSERT INTO '{sqliteProcessor.BaseIDPolicyTable}'(base_id, policies) VALUES('{sqliteProcessor.BaseID}', '{jsonTemplateObjectPolicy}')");
                                }
                                transaction.Commit();
                            }
                            catch (Exception ex)
                            {
                                Console.WriteLine(ex.ToString());
                                transaction.Rollback();
                            }
                            finally
                            {
                                sqliteProcessor.ExecuteAdmxNonQuery($"UPDATE BaseTable SET base_count = '{basecount}'");
                                transaction.Dispose();
                            }

                            progressBar.PercentComplete = (int)(countPercent * inc);
                            WriteProgress(progressBar);
                        }
                    }
                    sw.Stop();

                    WriteObject(new CreateDBInfo
                    {
                        dbFileLocation = new FileInfo(sqliteProcessor.Connection.DataSource).FullName,
                        createTime = sw.ElapsedMilliseconds.ToString()
                    });
                }
            }
        }
    }

    [Cmdlet(VerbsCommon.Join, "AdmxDatabases")]
    public class JoinAdmxDatabases : PSCmdlet
    {
        [Parameter(
            Mandatory = true,
            ValueFromPipelineByPropertyName = true,
            ValueFromPipeline = true,
            Position = 0,
            HelpMessage = "Join the tables from the source database to the destination database.\n INSERT source.sqlite3 INTO destination.sqlite3"
        )]
        public string Source { get; set; }

        [Parameter(
            Mandatory = true,
            ValueFromPipelineByPropertyName = true,
            ValueFromPipeline = true,
            Position = 1,
            HelpMessage = "Join the tables from the source database to the destination database.\n INSERT source.sqlite3 INTO destination.sqlite3"
        )]
        public string Destination { get; set; }


        protected override void ProcessRecord()
        {
            PluginResolver.InitPluginResolver(System.IO.Path.GetDirectoryName(System.Reflection.Assembly.GetExecutingAssembly().Location));
            using (SqliteProcessor sqliteProcessor = new SqliteProcessor())
            {
                sqliteProcessor.OpenAdmxSqliteDb(Destination);
                ProgressRecord progressBar = new ProgressRecord(1, "JoinAdmxDatabases", "Progress:");
                double inc = 0;
                double countPercent = (double)100 / (double)8;

                progressBar.PercentComplete = (int)(countPercent * inc);
                inc++;
                progressBar.StatusDescription = "Progress: ATTACH DATABASE Source";
                WriteProgress(progressBar);
                sqliteProcessor.ExecuteAdmxNonQuery($@"ATTACH DATABASE '{Source}' as sourcedb");

                string sourcedbBaseID = null;
                string sourcedbBaseIDPolicyTable = null;
                using (var command = sqliteProcessor.Connection.CreateCommand())
                {
                    command.CommandText = $"SELECT base_id from sourcedb.BaseTable";
                    var executeOutput = command.ExecuteReader();
                    while (executeOutput.Read())
                    {
                        sourcedbBaseID = executeOutput.GetString(0);
                    }
                }
                sourcedbBaseIDPolicyTable = sourcedbBaseID + "_BasePolicyTable";
                if (!String.IsNullOrEmpty(sourcedbBaseIDPolicyTable))
                {
                    progressBar.PercentComplete = (int)(countPercent * inc);
                    inc++;
                    progressBar.StatusDescription = "Progress: INSERT INTO Destination.BaseTable FROM Source.BaseTable";
                    WriteProgress(progressBar);
                    sqliteProcessor.ExecuteAdmxNonQuery($"INSERT INTO BaseTable SELECT * FROM sourcedb.BaseTable");

                    progressBar.PercentComplete = (int)(countPercent * inc);
                    inc++;
                    progressBar.StatusDescription = "Progress: CREATE TEMPORARY TABLE";
                    WriteProgress(progressBar);
                    sqliteProcessor.ExecuteAdmxNonQuery($@"CREATE TABLE IF NOT EXISTS '{sourcedbBaseIDPolicyTable}' ( [base_id] TEXT NOT NULL, [policies] TEXT NOT NULL, FOREIGN KEY (base_id) REFERENCES BaseTable (base_id) ON DELETE CASCADE )");

                    progressBar.PercentComplete = (int)(countPercent * inc);
                    inc++;
                    progressBar.StatusDescription = "Progress: INSERT INTO TEMPORARY TABLE FROM Destination.BasePolicyTable";
                    WriteProgress(progressBar);
                    sqliteProcessor.ExecuteAdmxNonQuery($@"INSERT INTO '{sourcedbBaseIDPolicyTable}' SELECT * FROM  sourcedb.'{sourcedbBaseIDPolicyTable}'");
                }
            }
        }
    }

    [Cmdlet(VerbsCommon.Get, "AdmxDatabaseInfo")]
    [OutputType(typeof(BaseTable))]
    public class GetAdmxDatabaseInfo : PSCmdlet
    {
        //[ValidatePattern("^([A-z]{1}:\\([A-z0-9.])+\\)[A-z0-9.]+")]
        [Parameter(
            Mandatory = true,
            ValueFromPipelineByPropertyName = true,
            ValueFromPipeline = true,
            Position = 0,
            HelpMessage = "The sqlite database source file from where the data should be read"
        )]
        public string Source { get; set; }

        protected override void ProcessRecord()
        {
            PluginResolver.InitPluginResolver(System.IO.Path.GetDirectoryName(System.Reflection.Assembly.GetExecutingAssembly().Location));
            using (SqliteProcessor sqliteProcessor = new SqliteProcessor())
            {
                sqliteProcessor.OpenAdmxSqliteDb(Source);
                using (var command = sqliteProcessor.Connection.CreateCommand())
                {
                    command.CommandText = $"SELECT * from BaseTable";
                    var executeOutput = command.ExecuteReader();
                    while (executeOutput.Read())
                    {
                        WriteObject(new BaseTable
                        {
                            base_id = executeOutput.GetString(0),
                            base_name = executeOutput.GetString(1),
                            base_time = executeOutput.GetString(2),
                            base_osversion = executeOutput.GetString(3),
                            base_count = executeOutput.GetInt32(4)
                        });
                    }
                }
            }
        }
    }

    [Cmdlet(VerbsCommon.Get, "AdmxDatabase")]
    [OutputType(typeof(BasePolicyTable))]
    public class GetAdmxDatabase : PSCmdlet
    {
        [Parameter(
            Mandatory = true,
            ValueFromPipelineByPropertyName = true,
            ValueFromPipeline = true,
            Position = 0,
            HelpMessage = "The sqlite database source file from where the data should be read."
        )]
        public string Source { get; set; }

        [Parameter(
            Mandatory = true,
            ValueFromPipelineByPropertyName = true,
            ValueFromPipeline = true,
            Position = 1,
            HelpMessage = "Join the tables from the source database to the destination database.\n INSERT source.sqlite3 INTO destination.sqlite3"
        )]
        public string BaseID { get; set; }

        protected override void ProcessRecord()
        {
            PluginResolver.InitPluginResolver(System.IO.Path.GetDirectoryName(System.Reflection.Assembly.GetExecutingAssembly().Location));
            if (!String.IsNullOrEmpty(BaseID))
            {
                using (SqliteProcessor sqliteProcessor = new SqliteProcessor())
                {
                    sqliteProcessor.OpenAdmxSqliteDb(Source);

                    using (var command = sqliteProcessor.Connection.CreateCommand())
                    {
                        command.CommandText = $@"SELECT * from '{BaseID + "_BasePolicyTable"}'";
                        var executeOutput = command.ExecuteReader();
                        while (executeOutput.Read())
                        {
                            WriteObject(new BasePolicyTable
                            {
                                base_id = executeOutput.GetString(0),
                                policies = executeOutput.GetString(1),
                            });
                        }
                    }
                }

            }
        }
    }

    [Cmdlet(VerbsCommon.Remove, "AdmxDatabase")]
    public class RemoveAdmxDatabase : PSCmdlet
    {
        [Parameter(
            Mandatory = true,
            ValueFromPipelineByPropertyName = true,
            ValueFromPipeline = true,
            Position = 1,
            HelpMessage = "The sqlite database source file from where the data should be deleted."
        )]
        public string Source { get; set; }

        [Parameter(
            Mandatory = true,
            ValueFromPipelineByPropertyName = true,
            ValueFromPipeline = true,
            Position = 0,
            HelpMessage = "path of admx database source."
        )]
        public string BaseID { get; set; }

        protected override void ProcessRecord()
        {
            PluginResolver.InitPluginResolver(System.IO.Path.GetDirectoryName(System.Reflection.Assembly.GetExecutingAssembly().Location));
            using (SqliteProcessor sqliteProcessor = new SqliteProcessor())
            {
                sqliteProcessor.OpenAdmxSqliteDb(Source);
                sqliteProcessor.ExecuteAdmxNonQuery($@"DELETE FROM BaseTable WHERE base_id = '{BaseID}'");
            }
        }
    }
}

